#!/usr/bin/env python

""" CLI entrypoint for pictorus device manager """
import argparse
from getpass import getpass
import platform
import socket
import sys
from urllib.parse import urljoin
import shutil
import os
from enum import Enum

import requests

from pictorus.config import API_PREFIX, Config
from pictorus.constants import PICTORUS_SERVICE_NAME
from pictorus.systemd import create_service, SYSTEMD_DIR

config = Config()


class TextFormat(Enum):
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"


def printf(message: str, fmt: TextFormat):
    print(f"{fmt.value}{message}{TextFormat.ENDC.value}")


def get_credentials():
    """Prompt user for username and password"""
    username = input("Enter username for pictorus: ")
    passwd = getpass("Enter password for pictorus: ")
    return username, passwd


def configure():
    """Configure the device manager"""
    configure_device()
    setup_device_manager()


def setup_device_manager():
    """Setup and start the device manager service"""
    print("Setting up device manager service")
    bin_path = shutil.which("pictorus-device-manager")
    if not bin_path:
        printf("Unable to set up device manager: executable missing", TextFormat.WARNING)
        return

    if not shutil.which("systemctl"):
        printf("Unable to set up device manager: systemctl not found", TextFormat.WARNING)
        return

    if not os.access(SYSTEMD_DIR, os.W_OK):
        printf(
            "Unable to set up device manager: permission denied\n"
            "Re-run as sudo or you can configure pictorus-device-manager to run manually",
            TextFormat.WARNING,
        )
        return

    create_service(
        PICTORUS_SERVICE_NAME,
        "Service to manage Pictorus apps",
        bin_path,
    )
    printf("Configured device manager service", TextFormat.OKGREEN)


def configure_device():
    """Configure this device to connect to pictorus"""
    print("Configuring new device")
    if not config.is_empty():
        confirm = input(
            "It looks like this device is already configured. Would you like to overwrite it [y/N]? "
        )
        if confirm.lower() != "y":
            printf("Skipping device registration", TextFormat.OKCYAN)
            return

    session = requests.Session()
    session.auth = get_credentials()

    hostname = socket.gethostname()
    device_name = input(f"Device Name [{hostname}]: ")
    device_name = device_name or hostname

    system_data = {
        "platform": platform.platform(),
        "system": platform.system(),
        "machine": platform.machine(),
    }

    res = session.post(
        urljoin(API_PREFIX, "v2/devices"),
        json={
            "name": device_name,
            "system": system_data,
        },
    )
    res.raise_for_status()
    config.store_config(res.json())
    printf(f"Successfully configured device: {device_name}", TextFormat.OKGREEN)


def main():
    """Main CLI entry point"""
    parser = argparse.ArgumentParser(description="Pictorus device manager")
    subparsers = parser.add_subparsers()

    parser_config = subparsers.add_parser("configure", help="Configure this device")
    parser_config.set_defaults(func=configure)

    args = parser.parse_args()

    if not hasattr(args, "func"):
        parser.print_help()
        sys.exit(0)

    args.func()


if __name__ == "__main__":
    main()
